# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


builtInFunctions = {
    "trace_vitis-ai-library": [
        "vitis::ai::ARFlowImp::run@libvitis_ai_library-arflow.so",
        "vitis::ai::BCCImp::bcc_post_process@libvitis_ai_library-bcc.so",
        "vitis::ai::BCCImp::preprocess@libvitis_ai_library-bcc.so",
        "vitis::ai::BCCImp::run@libvitis_ai_library-bcc.so",
        "vitis::ai::C2D2_liteImp::run@libvitis_ai_library-c2d2_lite.so",
        "vitis::ai::CarPlateRecogImp::run@libvitis_ai_library-carplaterecog.so",
        "vitis::ai::CenterPointImp::run@libvitis_ai_library-centerpoint.so",
        "vitis::ai::ClassificationImp::run@libvitis_ai_library-classification.so",
        "vitis::ai::ClocsImp::postprocess@libvitis_ai_library-clocs.so",
        "vitis::ai::ClocsImp::postprocess_kernel@libvitis_ai_library-clocs.so",
        "vitis::ai::ClocsImp::run@libvitis_ai_library-clocs.so",
        "vitis::ai::ClocsPointPillarsImp::postprocess@libvitis_ai_library-clocs.so",
        "vitis::ai::ClocsPointPillarsImp::run@libvitis_ai_library-clocs.so",
        "vitis::ai::ClocsPointPillarsImp::run_postprocess_t@libvitis_ai_library-clocs.so",
        "vitis::ai::ClocsPointPillarsImp::run_preprocess_t@libvitis_ai_library-clocs.so",
        "vitis::ai::ConfigurableDpuTaskImp::run@libvitis_ai_library-dpu_task.so",
        "vitis::ai::Covid19Segmentation8UC1::run@libvitis_ai_library-covid19segmentation.so",
        "vitis::ai::Covid19Segmentation8UC3::run@libvitis_ai_library-covid19segmentation.so",
        "vitis::ai::DetectImp::run@libvitis_ai_library-facedetect.so",
        "vitis::ai::DpuTaskImp::run@libvitis_ai_library-dpu_task.so",
        "vitis::ai::EfficientDetD2Imp::preprocess@libvitis_ai_library-efficientdet_d2.so",
        "vitis::ai::EfficientDetD2Imp::run@libvitis_ai_library-efficientdet_d2.so",
        "vitis::ai::FaceDetectRecogImp::run@libvitis_ai_library-facedetectrecog.so",
        "vitis::ai::FaceFeatureImp::run@libvitis_ai_library-facefeature.so",
        "vitis::ai::FaceLandmarkImp::run@libvitis_ai_library-facelandmark.so",
        "vitis::ai::FaceQuality5ptImp::run@libvitis_ai_library-facequality5pt.so",
        "vitis::ai::FaceRecogImp::run@libvitis_ai_library-facerecog.so",
        "vitis::ai::FairMotImp::run@libvitis_ai_library-fairmot.so",
        "vitis::ai::FusionCNNImp::postprocess@libvitis_ai_library-fusion_cnn.so",
        "vitis::ai::FusionCNNImp::preprocess@libvitis_ai_library-fusion_cnn.so",
        "vitis::ai::FusionCNNImp::run@libvitis_ai_library-fusion_cnn.so",
        "vitis::ai::HourglassImp::run@libvitis_ai_library-hourglass.so",
        "vitis::ai::MedicalDetectionImp::run@libvitis_ai_library-medicaldetection.so",
        "vitis::ai::MedicalSegcellImp::post_process@libvitis_ai_library-medicalsegcell.so",
        "vitis::ai::MedicalSegcellImp::run@libvitis_ai_library-medicalsegcell.so",
        "vitis::ai::MedicalSegmentationImp::run@libvitis_ai_library-medicalsegmentation.so",
        "vitis::ai::MnistClassificationImp::post_process@libvitis_ai_library-mnistclassification.so",
        "vitis::ai::MnistClassificationImp::pre_process@libvitis_ai_library-mnistclassification.so",
        "vitis::ai::MnistClassificationImp::run@libvitis_ai_library-mnistclassification.so",
        "vitis::ai::MultiTaskv3Imp::run_it@libvitis_ai_library-multitaskv3.so",
        "vitis::ai::OCRImp::preprocess@libvitis_ai_library-ocr.so",
        "vitis::ai::OCRImp::run@libvitis_ai_library-ocr.so",
        "vitis::ai::OFAYOLOImp::run@libvitis_ai_library-ofa_yolo.so",
        "vitis::ai::OpenPoseImp::run@libvitis_ai_library-openpose.so",
        "vitis::ai::PMGImp::pmg_post_process@libvitis_ai_library-pmg.so",
        "vitis::ai::PMGImp::run@libvitis_ai_library-pmg.so",
        "vitis::ai::PlateDetectImp::run@libvitis_ai_library-platedetect.so",
        "vitis::ai::PlateNumImp::run@libvitis_ai_library-platenum.so",
        "vitis::ai::PlateRecogImp::run@libvitis_ai_library-platerecog.so",
        "vitis::ai::PointPaintingImp::fusion@libvitis_ai_library-pointpainting.so",
        "vitis::ai::PointPaintingImp::run@libvitis_ai_library-pointpainting.so",
        "vitis::ai::PointPaintingImp::runPointPillars@libvitis_ai_library-pointpainting.so",
        "vitis::ai::PointPaintingImp::runSegmentation@libvitis_ai_library-pointpainting.so",
        "vitis::ai::PointPaintingImp::runSegmentationFusion@libvitis_ai_library-pointpainting.so",
        "vitis::ai::PointPillarsImp::run@libvitis_ai_library-pointpillars.so",
        "vitis::ai::PointPillarsNuscenesImp::run@libvitis_ai_library-pointpillars_nuscenes.so",
        "vitis::ai::PointPillarsNuscenesImp::sweepsFusionFilter@libvitis_ai_library-pointpillars_nuscenes.so",
        "vitis::ai::PointPillarsPost::post_process@libvitis_ai_library-pointpillars.so",
        "vitis::ai::PolypSegmentationImp::run@libvitis_ai_library-polypsegmentation.so",
        "vitis::ai::PoseDetectImp::run@libvitis_ai_library-posedetect.so",
        "vitis::ai::RGBDsegmentationImp::run@libvitis_ai_library-RGBDsegmentation.so",
        "vitis::ai::RcanImp::run@libvitis_ai_library-rcan.so",
        "vitis::ai::RefineDetImp::run@libvitis_ai_library-refinedet.so",
        "vitis::ai::ReidImp::run@libvitis_ai_library-reid.so",
        "vitis::ai::RetinaFaceImp::run@libvitis_ai_library-retinaface.so",
        "vitis::ai::RoadLineImp::run@libvitis_ai_library-lanedetect.so",
        "vitis::ai::SSDImp::run@libvitis_ai_library-ssd.so",
        "vitis::ai::Segmentation3DImp::run@libvitis_ai_library-3Dsegmentation.so",
        "vitis::ai::Segmentation3DPost::post_prec@libvitis_ai_library-3Dsegmentation.so",
        "vitis::ai::SoloImp::run@libvitis_ai_library-solo.so",
        "vitis::ai::TFSSDImp::run@libvitis_ai_library-tfssd.so",
        "vitis::ai::TextMountainImp::run@libvitis_ai_library-textmountain.so",
        "vitis::ai::UltraFastImp::run@libvitis_ai_library-ultrafast.so",
        "vitis::ai::VehicleClassificationImp::run@libvitis_ai_library-vehicleclassification.so",
        "vitis::ai::YOLOv2Imp::run@libvitis_ai_library-yolov2.so",
        "vitis::ai::YOLOv3Imp::run@libvitis_ai_library-yolov3.so",
        "vitis::ai::YOLOvXImp::run@libvitis_ai_library-yolovx.so",
        "vitis::ai::bev_preprocess@libvitis_ai_library-pointpillars.so",
        "vitis::ai::centerpoint::middle_process@libvitis_ai_library-centerpoint.so",
        "vitis::ai::centerpoint::preprocess3@libvitis_ai_library-centerpoint.so",
        "vitis::ai::fairmot_post_process@libvitis_ai_library-fairmot.so"
    ],
    "trace_xnnpp_post_process": [
        "vitis::ai::EfficientDetD2Post::postprocess@libxnnpp-xnnpp.so",
        "vitis::ai::EfficientDetD2Post::postprocess_kernel@libxnnpp-xnnpp.so",
        "vitis::ai::MedicalSegmentationPost::medicalsegmentation_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::MultiTaskPostProcessImp::post_process_seg@libxnnpp-xnnpp.so",
        "vitis::ai::MultiTaskPostProcessImp::process_det@libxnnpp-xnnpp.so",
        "vitis::ai::MultiTaskPostProcessImp::process_seg@libxnnpp-xnnpp.so",
        "vitis::ai::MultiTaskPostProcessImp::process_seg_visualization@libxnnpp-xnnpp.so",
        "vitis::ai::MultiTaskv3PostProcessImp::post_process@libxnnpp-xnnpp.so",
        "vitis::ai::MultiTaskv3PostProcessImp::post_process_visualization@libxnnpp-xnnpp.so",
        "vitis::ai::OCRPostImp::process@libxnnpp-xnnpp.so",
        "vitis::ai::PointPillarsNuscenesPost::postprocess@libxnnpp-xnnpp.so",
        "vitis::ai::RefineDetPost::refine_det_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::RetinaFacePost::retinaface_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::RoadLinePost::road_line_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::SSDPost::ssd_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::TFSSDPost::ssd_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::TextMountainPostImp::process@libxnnpp-xnnpp.so",
        "vitis::ai::UltraFastPostImp::post_process@libxnnpp-xnnpp.so",
        "vitis::ai::X_Autonomous3DPost::process@libxnnpp-xnnpp.so",
        "vitis::ai::classification_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::face_detect_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::face_feature_post_process_fixed@libxnnpp-xnnpp.so",
        "vitis::ai::face_feature_post_process_float@libxnnpp-xnnpp.so",
        "vitis::ai::face_landmark_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::face_quality5pt_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::hourglass_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::medicaldetection::MedicalDetectionPost::medicaldetection_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::ofa_yolo_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::open_pose_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::plate_detect_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::plate_num_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::pose_detect_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::post_process@libxnnpp-xnnpp.so",
        "vitis::ai::process_kernel@libxnnpp-xnnpp.so",
        "vitis::ai::rcan_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::reid_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::segmentation_post_process_8UC1@libxnnpp-xnnpp.so",
        "vitis::ai::segmentation_post_process_8UC3@libxnnpp-xnnpp.so",
        "vitis::ai::solo_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::solo_post_process_batch@libxnnpp-xnnpp.so",
        "vitis::ai::tfrefinedet::TFRefineDetPost::tfrefinedet_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::vehicleclassification_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::yolov2_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::yolov3_post_process@libxnnpp-xnnpp.so",
        "vitis::ai::yolovx_post_process@libxnnpp-xnnpp.so"
    ],
    "trace_vart": [
        "convert_data_type@libvart-runner.so",
        "convert_tensors@libvart-runner.so",
        "vart::TensorBuffer::copy_from_host@libvart-runner.so",
        "vart::TensorBuffer::copy_tensor_buffer@libvart-runner.so",
        "vart::TensorBuffer::copy_to_host@libvart-runner.so",
        "xir::XrtCu::run@libvart-dpu-controller.so"
    ],
    "trace_opencv": [
        "cv::imread",
        "cv::imshow",
        "cv::resize"
    ]
}

trace_va_timeout = 60
trace_xat_timeout = 60
trace_max_timeout = 120
trace_fg_timeout = 30
trace_max_fg_timeout = 60

default_runmode = "normal"
#default_runmode = "debug"

traceCfgDefaule = {
    'collector': {},
    'tracer': {
        'power': {
            'disable': True,
        },
        'sched': {
            'disable': True,
        },
    },
    'trace': {
        "enable_trace_list": ["vitis-ai-library", "vart", "xnnpp_post_process", "opencv", "custom"]
    },
    'control': {
        'cmd': None,
        'xat': {
            'compress': True,
            'filename': None
        },
        'config': None,
        'timeout': None,
    }
}



# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .ascii_table import *
from collections import namedtuple

import sys
import os
import logging

from writer.parser.vart_parser import *
from writer.parser.tracepointUtil import *
from writer.parser.statUtil import *
from writer.parser.pyfunc import *
from writer.parser.cppfunc import *
from writer.parser.cpu_task import *

vtf_events = []
literal_pool = []
dpu_core_num = 0
DEBUG_MODE = False
XRT_INFO = {}
TIMESYNC = {}
dpu_profile_summary = []
cpuTaskSummary = []
cppFuncSummary = []
pyFuncSummary = []
SUBG_DPSP_LEN = 60
OPS_DPSP_LEN = 60

output_f = sys.stdout


def convert_dpu(raw_data, hwInfo, options):
    global literal_pool, dpu_core_num, vtf_events, dpu_profile_summary
    global TIMESYNC
    offset = TIMESYNC.get('sync', {}).get('vart')
    timeout = TIMESYNC.get('timeout', 0)

    runmode = options.get('control', {}).get('runmode')

    runmachine = options.get('control', {}).get('platform',{}).get('machine',{})
    dpu_parser = DPUEventParser()
    timelines = dpu_parser.parse(
            raw_data, hwInfo, {"time_offset": offset, "time_limit": timeout, "runmode": runmode})

    dpu_profile_summary = dpu_parser.get_dpu_profile_summary( {"run_machine": runmachine}, "txt")

    """extracting all strings"""
    for dpu in timelines:
        for event in dpu.timeline:
            literal_pool.append(event.subgraph)
            literal_pool.append(event.it)
            literal_pool.append(event.ot)
    literal_pool = list(set(literal_pool))
    literal_pool.insert(0, "RESERVED")

    for dpu_timeline in timelines:
        if len(dpu_timeline.timeline) == 0:
            continue
        dpu_core_num += 1
        for event in dpu_timeline.timeline:
            pass

    vtf_events.sort(key=lambda x: x.ts)


DPU_TABLE_NOTES = \
    """
"~0": Value is close to 0, Within range of (0, 0.001)
Bat: Batch size of the DPU instance
WL(Work Load): Computation workload (MAC indicates two operations), unit is GOP
SW_RT(Software Run time): The execution time calculate by software in milliseconds, unit is ms
HW_RT(Hareware Run time): The execution time from hardware counter in milliseconds, unit is ms
Effic(Efficiency): The DPU actual performance divided by peak theoretical performance,unit is %
LdFM(Load Size of Feature Map): External memory load size of feature map, unit is MB
LdWB(Load Size of Weight and Bias): External memory load size of bias and weight, unit is MB
StFM(Store Size of Feature Map): External memory store size of feature map, unit is MB
AvgBw(Average bandwidth): External memory average bandwidth. unit is MB/s
....
"""

DPU_TABLE_CLOUD_NOTES = \
    """
"~0": Value is close to 0, Within range of (0, 0.001)
Bat: Batch size of the DPU instance
WL(Work Load): Computation workload (MAC indicates two operations), unit is GOP
SW_RT(Software Run time): The execution time calculate by software in milliseconds, unit is ms
LdFM(Load Size of Feature Map): External memory load size of feature map, unit is MB
LdWB(Load Size of Weight and Bias): External memory load size of bias and weight, unit is MB
StFM(Store Size of Feature Map): External memory store size of feature map, unit is MB
AvgBw(Average bandwidth): External memory average bandwidth. unit is MB/s
....
"""

def print_dpu_table(dpu_summary_data, runmachine):
    """
    ['subgraph_conv1,651,DPUCZDX8G_1:batch-1,0.067,11.461,34.191,7.718,673.388,49.947,4357.893,\n',
     'subgraph_conv1,215,DPUCZDX8G_2:batch-1,34.404,34.585,45.104,7.718,223.155,49.947,1444.167,\n']
    """
    #gops = hwin_summary_data[1]
    #freq = hwin_summary_data[0]
    header = ['DPU Id', 'Bat', 'DPU SubGraph',
              'WL', 'SW_RT', 'HW_RT', 'Effic', 'LdWB', 'LdFM', 'StFM', 'AvgBw']

    header_cloud = ['DPU Id', 'Bat', 'DPU SubGraph',
                          'WL', 'SW_RT', 'LdWB', 'LdFM', 'StFM', 'AvgBw']
    pr_data = []
    # pr_data.append(header)

    for row in dpu_summary_data:
        items = row.strip().split(',')
        subgraph_name = items[0].replace("subgraph_", "", 1)
        if len(subgraph_name) > SUBG_DPSP_LEN:
            subgraph_name = "..." + subgraph_name[-SUBG_DPSP_LEN:]

        runs = items[1]

        device_raw = items[2]
        ip_name = device_raw.split(':')[0]
        ip_batch = device_raw.split(':')[1].split('-')[1]

        min_t = items[3]
        ave_t = items[4]
        max_t = items[5]

        workload = items[6]
        effic = items[7]
        mem_ld_fm = items[8]
        mem_ld_w = items[9]
        mem_st_fm = items[10]
        mem_io_bw = items[11]

        hw_rt = items[12]
        rank_id = int(items[13])
        #bandwidth = "{:,}".format(round(float(items[11])))
        if runmachine == "aarch64":
            pr_data.append([ip_name, ip_batch, subgraph_name,
                workload, ave_t, hw_rt, effic, mem_ld_fm, mem_ld_w, mem_st_fm, mem_io_bw, rank_id])
        elif runmachine == "x86_64":
            pr_data.append([ip_name, ip_batch, subgraph_name,
                workload, ave_t, mem_ld_fm, mem_ld_w, mem_st_fm, mem_io_bw, rank_id])

        

    pr_data.sort(key=lambda a: (a.pop() + (hash(a[0]) % 128) * 4096))

    if len(pr_data) == 0:
        return
    if runmachine == "aarch64":
        pr_data.insert(0, header)
    elif runmachine == "x86_64":
        pr_data.insert(0, header_cloud)
    print("DPU Summary:", file=output_f)
    print_ascii_table(pr_data, output_f)
    if runmachine == "aarch64":
        print("\nNotes:%s" % DPU_TABLE_NOTES, file=output_f)
    elif runmachine == "x86_64":
        print("\nNotes:%s" % DPU_TABLE_CLOUD_NOTES, file=output_f)


def print_cpu_task_table(cpu_task_summary):
    if len(cpu_task_summary) == 0:
        return

    header = ['CPU SubGraph', 'OPs', 'Device', 'Runs', 'AverageRunTime(ms)']
    pr_data = []
    pr_data.append(header)

    for row in cpu_task_summary:
        items = row.strip().split(',')

        subgraph_name = items[0].replace("subgraph_", "", 1)
        if len(subgraph_name) > SUBG_DPSP_LEN:
            subgraph_name = "..." + subgraph_name[-SUBG_DPSP_LEN:]
        ops = ""
        dev = "CPU"

        runs = items[1]
        min_t = items[3]
        ave_t = items[4]
        max_t = items[5]
        ops = items[6]
        if (len(ops) > OPS_DPSP_LEN):
            ops = "..." + ops[-OPS_DPSP_LEN:]

        pr_data.append([subgraph_name, ops, dev, runs, ave_t])

    print("\nCPU OPs in Graph(called by GraphRunner):", file=output_f)
    print_ascii_table(pr_data, output_f)


def print_power_info_table(power_info_data):

    header = ['idle power', 'max power', 'ave power']
    pr_data = []
    pr_data.append(header)

    idle_power = power_info_data[0]
    max_power = power_info_data[1]
    ave_power = power_info_data[2]

    pr_data.append([idle_power, max_power, ave_power])

    print("power summary:", file=output_f)
    print_ascii_table(pr_data, output_f)


def print_cpu_func_table(cpp_summary, py_summary):
    """
    ['vitis::ai::ConfigurableDpuTaskImp::setInputImageBGR_1,657,CPU,0.474,0.488,0.652,\n',
     'xir::XrtCu::run,1314,CPU,0.063,7.181,21.457,\n', 'vitis::ai::DpuTaskImp::run,657,CPU,13.180,14.381,21.618,\n']
    """

    header = ['Function', 'Device', 'Runs', 'AverageRunTime(ms)']
    pr_data = []
    pr_data.append(header)

    cpu_func_summary = cpp_summary + py_summary

    if len(cpu_func_summary) == 0:
        return

    for row in cpp_summary:
        items = row.strip().split(',')

        function_name = items[0]
        dev = "CPU"
        runs = items[1]

        min_t = items[3]
        ave_t = items[4]
        max_t = items[5]

        pr_data.append([function_name, dev, runs, ave_t])

    for row in py_summary:
        items = row.strip().split(',')

        function_name = "%s@py" % items[0]
        dev = "CPU"
        runs = items[1]

        min_t = items[3]
        ave_t = items[4]
        max_t = items[5]

        pr_data.append([function_name, dev, runs, ave_t])

    print("\nCPU Functions(Not in Graph, e.g.: pre/post-processing, vai-runtime):", file=output_f)
    print_ascii_table(pr_data, output_f)


# Statistical information for DPU kernels(min/max time etc.)
def output_profile_summary(runmachine):
    global pyFuncSummary, cpuTaskSummary, cppFuncSummary, dpu_profile_summary

    print_dpu_table(dpu_profile_summary, runmachine)
    print_cpu_task_table(cpuTaskSummary)
    print_cpu_func_table(cppFuncSummary, pyFuncSummary)


def output(runmachine, saveTo=None):
    global output_f

    if saveTo != None:
        logging.info("Saving report to: %s" % saveTo)
        try:
            output_f = open(saveTo, "w+t")
        except:
            output_f = sys.stdout
            logging.error("Fail opening: %s" % saveTo)

    output_profile_summary(runmachine)


def xat_to_txt(xat, options):
    global XRT_INFO, DEBUG_MODE, cpuTaskSummary, cppFuncSummary, pyFuncSummary

    saveTo = options.get('cmdline_args', {}).get('output', None)
    runmachine = options.get('control', {}).get('platform',{}).get('machine',{})

    convert_dpu(xat.get('vart'), xat.get('hwInfo'), options)
    cpuTaskSummary = convert_cpu_task(xat.get('vart', {}))
    pyFuncSummary = convert_pyfunc(xat.get('pyfunc', {}))
    cppFuncSummary = convert_cppfunc(xat.get('function', {}))

    output(runmachine,saveTo)




# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def merge(a: dict, b: dict):
    if hasattr(a, "keys") and hasattr(b, "keys"):
        for kb in b.keys():
            if kb in a.keys():
                merge(a[kb], b[kb])
            else:
                a.update(b)


def get(opt, key: str):
    t = opt.copy()
    for k in key.split('.'):
        t = t.get(k, {})
    return t


def set(opt, key: str, value):
    for k in key.split('.')[0:-1]:
        if not k in opt.keys():
            opt[k] = {}
        opt = opt[k]
        print(key.index(k), "### key:", k, " v:", opt)

    opt.update({key[-1]: value})



#!/usr/bin/python3
# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import platform

tmp_dir = "/workspace/tmp/"
tmp_lib_dir = os.path.join(tmp_dir, "lib")
ori_lib_dir = "/usr/lib/."

user_cmd = ""
for argv in sys.argv[1:]:
    user_cmd += " %s" % argv
vaitrace_cmd = "vaitrace"
sudo_cmd = "sudo -E"
env_cmd = ""

# 1. Env check, arch & overlayfs


def overlay_fs_bug_exist():
    pl = platform.uname().release.split('.')
    kernel_maj, kernel_min = int(pl[0]), int(pl[1])
    if kernel_maj >= 4 and kernel_min >= 8:
        return False

    overlay_fs = os.system('o=`df /usr/lib/ |grep overlay`;test X\"$o\" = X')
    if overlay_fs == 0:
        return False
    print("overlay_fs found! work it round")
    return True

# if inside_docker:
#  ...
# else:
#  ...

# check pth [/sys/kernel/debug/tracing]


def check_ftrace_path():
    return True


def overlay_fs_workaround():
    # 1. Create temp dir and copy libs
    os.system("mkdir -p %s" % tmp_lib_dir)
    print("Copying libraries")
    os.system("cp -r %s %s" % (ori_lib_dir, tmp_lib_dir))
    print("Copy finished")

    # 2. Change $LD_LIBRARY_PATH
    LD_LIBRARY_PATH = os.getenv("LD_LIBRARY_PATH")
    print(LD_LIBRARY_PATH)
    NEW_LD_LIBRARY_PATH = tmp_lib_dir + ':' + LD_LIBRARY_PATH
    print("new LD_LIBRARY_PATH:\n", NEW_LD_LIBRARY_PATH)
    global env_cmd
    env_cmd = 'env LD_LIBRARY_PATH="%s"' % NEW_LD_LIBRARY_PATH


def clean_tmp():
    print("Removing temp files")
    os.system("rm -rf %s" % tmp_dir)

# 4. Issue the command with sudo -E env $LD_LIBRARY_PATH $PATH $LIBRARY_PATH

# 5. Change mode of current dir for *.csv files


def chmod_cur():
    print("Changing mode of current dir")
    os.system("chmod 777 ./")


if not check_ftrace_path():
    print("Checking ftrace failed")

if overlay_fs_bug_exist():
    overlay_fs_workaround()
    chmod_cur()


exec_cmd = "%s %s %s %s" % (sudo_cmd, env_cmd, vaitrace_cmd, user_cmd)
print(exec_cmd)
os.system(exec_cmd)
clean_tmp()



#!/usr/bin/python3
# -*- coding: UTF-8 -*-

# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import re
import csv
import sys
import string
import time
import argparse
import json
import signal
import logging
from subprocess import Popen, PIPE

import vaitraceDefaults


def merge(a: dict, b: dict):
    if hasattr(a, "keys") and hasattr(b, "keys"):
        for kb in b.keys():
            if kb in a.keys():
                merge(a[kb], b[kb])
            else:
                a.update(b)


def saityCheck(options: dict):
    cmd = options['control']['cmd']
    if cmd == None or len(cmd) == 0:
        logging.error("Wrong command")
        return False

    timeout = options['control']['timeout']
    if timeout < 0:
        logging.error("Wrong timeout")
        return False

    output = options['control']['xat']['filename']
    if output == None or output == "":
        logging.errer("Wrong output file path")
        return False

    logging.debug("Control Options:")
    logging.debug(options['control'])

    return True


def parseCfg(options: dict):

    cfg_path = options.get('cmdline_args').get('config')
    if cfg_path != "":
        try:
            cfg_file = open(cfg_path, 'rt')
            cfg = json.load(cfg_file)
            overlayOption = cfg.get('options', {})
        except:
            logging.error(f"Invalid config file: {cfg_path}")
            exit(-1)
        logging.info(f"Applying Config File {cfg_path}")
    else:
        cfg = options
        overlayOption = {}

    """Configuration priority: Configuration File > Command Line > Default"""
    """Cmd"""
    cmdline_cmd = options['cmdline_args'].get("cmd", "")
    cmd = overlayOption.pop("cmd", cmdline_cmd)
    if cmd == None or len(cmd) == 0:
        logging.error("Can not find excutable command")
        return False

    if type(cmd) == str and cmd != "":
        cmd = cmd.split()
    options['control']['cmd'] = cmd

    """Output"""
    cmdline_output = options['cmdline_args'].get("output", "")
    output = overlayOption.pop("output", cmdline_output)
    if output == "" or output == None:
        shortcmd = options['control']['cmd'][0].split('/')[-1][:15]
        output = os.path.join(os.path.abspath(os.curdir), "%s.xat" % shortcmd)

    options['control']['xat']['filename'] = output

    """Timeout"""
    cmdline_timeout = options['cmdline_args'].get("timeout")
    timeout = overlayOption.pop("timeout", cmdline_timeout)
    options['control']['timeout'] = timeout

    """Vitis Analyzer Mode"""
    """
    For Vitis Analyzer:
    1. do a long trace (30s)
    2. disable [sched] tracer
    3. set env[Debug.xrt_profile] and [Debug.vitis_ai_profile] for xrt
    """
    va_enabled = options['cmdline_args']['va']
    if va_enabled:
        if options['control']['timeout'] == None:
            options['control']['timeout'] = vaitraceDefaults.trace_va_timeout
        # merge(options, {'tracer': {'sched': {'disable': True}}})
        """xrt fixed it, [Debug.xrt_profile] is no longger required for v1.4, but 2020.2 still need it"""
        os.environ.setdefault("Debug.xrt_profile", "true")
        os.environ.setdefault("Debug.vitis_ai_profile", "true")
    else:
        if options['control']['timeout'] == None:
            options['control']['timeout'] = vaitraceDefaults.trace_xat_timeout

    """Fine grained"""
    fg = options['cmdline_args'].get('fg', False)
    if fg:
        options['control']['timeout'] = vaitraceDefaults.trace_fg_timeout

    """Debug"""
    cmdline_debug = options['cmdline_args'].get('debug', False)
    debug = overlayOption.pop("debug", cmdline_debug)
    options['control']['debug'] = debug
    if (debug):
        os.environ.setdefault("DEBUG_VAITRACE", "1")

    """TraceList"""
    traceList = {}
    traceCfg = cfg.get('trace', {})

    builtInList = vaitraceDefaults.builtInFunctions
    defaultEnabledTraceList = vaitraceDefaults.traceCfgDefaule['trace']["enable_trace_list"]
    enabledTraceList = traceCfg.get(
        'enable_trace_list', defaultEnabledTraceList)
    for trace in enabledTraceList:
        traceName = "trace_" + trace
        if (traceName) in builtInList.keys():
            traceList.update({traceName: builtInList[traceName]})

    traceList.update({"trace_custom": cfg.get("trace_custom", [])})

    merge(options, {'tracer': {'function': {'traceList': traceList}}})

    """Other Options"""
    merge(options, overlayOption)

    return True



#!/usr/bin/python3

# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys


def print_ascii_table(rows, saveTo=sys.stdout):

    output_ascii_table = saveTo
    headers = tuple(rows[0])
    headers_width = len(headers)

    lens = []

    def str_len(_x):
        return len(str(_x))

    for line in rows[1:]:
        line_width = len(line)
        if (line_width > headers_width):
            line = line[0:headers_width]
        if (line_width < headers_width):
            len_diff = headers_width - line_width
            line.extend(list(len_diff * '-'))

    for i in range(len(headers)):
        lens.append(
            str_len(max([x[i] for x in rows], key=lambda x: str_len(x))))

    formats = []
    hformats = []
    for i in range(len(headers)):
        formats.append("%%-%ds" % lens[i])
        hformats.append("%%-%ds" % lens[i])
    pattern = " | ".join(formats)
    hpattern = " | ".join(hformats)
    separator = "-+-".join(['-' * n for n in lens])

    table_width = len(hpattern % headers)
    print('=' * table_width, file=output_ascii_table)
    print(hpattern % headers, file=output_ascii_table)
    print(separator, file=output_ascii_table)

    def _u(t): return t
    for line in rows[1:]:
        print(pattern % tuple(_u(t) for t in line), file=output_ascii_table)

    print('=' * table_width, file=output_ascii_table)


class ascii_table:
    def __init__(_headers, _title=""):
        title = _title
        header = _headers
        data = []
        row_num = 1
        column_num = len(header)
        column_lens = []

        pass

    def add_data():
        pass

    def print():
        table_width = 0
        pass


if __name__ == "__main__":
    titles = ['SubGraph', 'Workload(GOPS)', 'AverageRunTime(ms)', 'Perf(GOPS)']
    data_1 = ["conv1",          123.333,    99.88, '694.2']
    data_2 = ['res2a_branch2a', 3129.099,   124.4, '-']
    data_3 = ["res2a_branch2b", 938.33,     32.33, '1101.0']
    data_4 = ["res2a_branch2c", 232.22,    32.3,  '676.8']
    data_5 = ["res3c_branch2c", 102.760,   33.5,  '556.6']

    """test rows"""
    print_ascii_table([titles, data_1, data_2, data_3, data_4, data_5])




# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto


class vaiTimelineEventState(Enum):
    START = auto()
    DONE = auto()
    PERIOD = auto()
    MARKER = auto()


class vaiTimelineEvent:
    def __str__(self):
        if self.type == 'period':
            t = "Duration:%10.8f" % self.duration
        else:
            t = "TimeStamp:%10.8f" % self.ts

        return "Core %10s-%02d: Type:%6s, pid:%06d ts:%f" % \
            (self.coreType, self.coreId, self.type, self.pid, self.ts) + " " + t

    """
    Four types: start, end (the same with done), marker, and period
    """

    def __init__(self, _ts, _pid, _eventType, _coreType, _coreId, _info=None):
        self.ts = float(_ts)
        self.pid = int(_pid)
        self.type = _eventType
        self.coreType = _coreType
        self.coreId = int(_coreId)
        self.startTime = 0
        self.endTime = 0
        self.duration = 0
        self.info = _info

        if self.type == 'start' or self.type == 'done':
            self.code = _info

    def paired(self, end):
        if not (self.type == 'start' and end.type == 'done'):
            return None
        if end.ts < self.ts:
            return None

        self.type = 'period'
        self.startTime = self.ts
        self.endTime = end.ts
        self.duration = self.endTime - self.startTime

    def get_info(self, key: str):
        return self.info.get(key, None)


def dpuEventPaired(start: vaiTimelineEvent, end: vaiTimelineEvent):
    if not (start.type == 'start' and end.type == 'done'):
        return None
    if end.ts < start.ts:
        return None
    if start.pid != end.pid:
        return None
    if start.coreId != end.coreId:
        return None

    start.type = 'period'
    start.startTime = start.ts
    start.endTime = end.ts
    start.duration = start.endTime - start.startTime
    start.hwcounter = int(end.info.get("hwconuter",0))

    return start


class vaiTimeline:
    def __init__(self, _coreType, _coreId: str, options: dict, _timeout: int = 0):
        self.coreType = _coreType
        self.coreId = _coreId
        self.dpuCore = None
        self.lastTraceItem = None
        self.timeline = []
        self.eventStack = []
        self.timeout = _timeout
        self.transTs = lambda x: round(float(x), 6)
        if options.get('traceClock', "") == 'x86-tsc':
            tsc_hz = int(options.get('x86_tsc_khz') * 1000)
            self.transTs = lambda x: round((float(x) / tsc_hz), 6)

    def tracsTs_x86_tsc(self):
        pass

    def get_util(self, _filter=None):
        """Drop all marker events"""
        events = [e for e in self.timeline if e.type == 'period']

        if _filter is not None:
            events = [e for e in events if _filter(e)]

        if len(events) == 0:
            return -1

        totalT = events[-1].endTime - events[0].startTime
        runT = 0

        for e in events:
            runT += e.duration

        util = runT * 100 / totalT
        # print("### Util of %s-%02d: %.2f" % (self.coreType, self.coreId, util))

        return util

    def get_core_name(self):
        try:
            n = self.dpuCore.name
        except:
            n = None

        return n

    def get_core_full_name(self):
        try:
            n = self.dpuCore.full_name
        except:
            n = None

        return n

    def len(self):
        return len(self.timeline)

    def add(self, new: vaiTimelineEvent):
        if new.coreId != self.coreId:
            return

        new.ts = self.transTs(new.ts)

        if new.type == 'start':
            self.eventStack.append(new)
        if new.type == 'done':
            """1. looking for start event in eventstack"""
            for s in self.eventStack[::-1]:
                r = dpuEventPaired(s, new)
                if r != None:
                    try:
                        self.eventStack.remove(s)
                    except BaseException:
                        pass
                    """2. if matched, add into self.timeline"""
                    self.timeline.append(r)

    def toJson(self, prefix="thread", _color=None):
        out = []
        for i in self.timeline:
            if i.type != 'period':
                continue

            if _color is None or _color(i) is None:
                color = "#%06x" % ((hash(str(i.pid)) >> 2) % (256 * 256 * 256))
            else:
                color = _color(i)

            if i.code != "":
                prefix = i.code
            out.append([prefix, i.pid, i.startTime, i.endTime, color])

        return out


def createTimelines(coreType, id, options: dict, timeout=0):
    timelines = []

    if isinstance(id, list):
        for i in id:
            timelines.append(vaiTimeline(coreType, i, options, timeout))
    elif isinstance(id, str):
        timelines.append(vaiTimeline(coreType, id, options, timeout))
    elif isinstance(id, int):
        for i in range(0, id):
            timelines.append(vaiTimeline(coreType, i, options, timeout))

    if len(timelines) == 0:
        assert ()

    return timelines



#!/usr/bin/python3
# -*- coding:utf-8 -*-

# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
import os
import logging
import time
import math
import copy
import threading
from ctypes import *
import copy
import logging
import json

idcode_ddrmc_num_dict = {'0x4c9b093':3, '0x4c93093':2, '0x4c98093':3, '0x4ca9093':4, '0x4ca8093':4, '0x4cd2093':3, '0x4cd0093':3,
                 '0x4c9a093':3, '0x4cc1093':1, '0x4cc0093':1, '0x4cc9093':1, '0x4cc8093':1, '0x4cd2093':3, '0x4cd3093':3}

def merge(a: dict, b: dict):
    if hasattr(a, "keys") and hasattr(b, "keys"):
        for kb in b.keys():
            if kb in a.keys():
                merge(a[kb], b[kb])
            else:
                a.update(b)

class NOCRecord(Structure):
    _fields_ = [('time', c_double), ('data', c_ulonglong * 10)]

    def output(self):
        return "NOC %.7f %d %d %d %d %d %d %d %d %d %d\n" % (self.time, *self.data)


class NOC:
    def __init__(self, mem_type="noc", path='/usr/lib/libmemperf.so', enable=True):
        self.nocLib = cdll.LoadLibrary(path)
        self.data = []
        self.base_addr = []
        self.ddrmc_name = []
        self.interval = 0
        self.act_period = 0
        self.record_data_len = 0
        self.type = mem_type
        self.enabled = enable
        self.idcode = self.getChipId()
        self.get_noc_base_addr(self.idcode)
        self.noc_base_addr = (c_int * len(self.base_addr))(*self.base_addr)
        self.nocLib.create_noc_instance(
            self.type.encode(), self.noc_base_addr, int(len(self.base_addr)))

    def start(self, interval=0.01):
        if self.enabled == False:
            return
        self.interval = interval
        self.nocLib.start.restype = c_int
        return self.nocLib.start(c_double(interval))

    def stop(self):
        if self.enabled == False:
            return

        self.nocLib.stop()
        data = NOCRecord()
        pd = pointer(data)

        while (self.nocLib.pop_data(pd) == 0):
            self.data.append(copy.deepcopy(data))

        self.nocLib.get_act_period.restype = c_double
        self.act_period = self.nocLib.get_act_period()

        self.nocLib.get_record_data_len.restype = c_int
        self.record_data_len = self.nocLib.get_record_data_len()

    def transTimebase(self):
        if self.enabled == False:
            return
        for i in range(0, len(self.data)):
            for j in range(0, self.record_data_len):
                self.data[i].data[j] = int(
                    self.data[i].data[j] / self.interval)

    def printData(self, data):
        if self.enabled == False:
            return
        info = str()
        info += "TimeStamp: %.7f\n" % data.time

        read = [data.data[i]/self.interval/1000 /
                1000 for i in range(0, 10) if i % 2 == 0]
        info += "Read Ports:  "
        for d in read:
            info += "%8.1f" % d

        info += " MB/s\n"

        write = [data.data[i]/self.interval/1000 /
                 1000 for i in range(0, 10) if i % 2 == 1]
        info += "Write Ports: "
        for d in write:
            info += "%8.1f" % d

        info += " MB/s\n  "

        return info
    
    
    def get_noc_base_addr(self, devid):
        ddrmc_number = idcode_ddrmc_num_dict.get(devid, 0)
        if ddrmc_number == 0:
            logging.info("Can not get ddrmc/noc number!")
            return
        pre_base_addr = []
        for (dirpath, dirnames, filenames) in os.walk("/proc/device-tree/axi"):
            for file in dirnames:
                if file.startswith("memory-controller"):
                    compatible = open(os.path.join(os.path.abspath(os.path.join(
                        dirpath, file)), "compatible"), "rt").read().strip().split(',')
                    driver_name = compatible[1].split('-')
                    if driver_name[1] != 'ddrmc':
                        continue
                    reg = os.path.join(os.path.abspath(
                        os.path.join(dirpath, file)), "reg")
                    f = open(reg, "rb")
                    numl = list(f.read())
                    addr_off = ((numl[20] << 24) + (numl[21] <<
                                16) + (numl[22] << 8) + (numl[23] << 0))
                    pre_base_addr.append(int(addr_off))
        pre_base_addr.sort()
        self.base_addr =  pre_base_addr[0:ddrmc_number]
    
    def getChipId(self):
        try:
            assert (os.access("/sys/kernel/debug/zynqmp-firmware/pm", os.W_OK) == True)
            pm = open("/sys/kernel/debug/zynqmp-firmware/pm", "wt")
            assert (pm.write("PM_GET_CHIPID") == len("PM_GET_CHIPID"))
            pm.close()
            pm = open("/sys/kernel/debug/zynqmp-firmware/pm", "rt")
            chip_id_raw = pm.read().strip()  # 'Idcode: 0x4724093, Version:0x20000513'
            pm.close()
            idcode = (chip_id_raw.split(',', 1)[0]).split(':', 1)[1].strip()
            return idcode
        except:
            logging.error(
                    "Cannot get chip id, kernel config CONFIG_ZYNQMP_FIRMWARE_DEBUG=y is required.")
            exit(1)
    
    def getData(self):
        return [d.output() for d in self.data]

    def toMB_s(self,r):
        # 100 M 
        return float(r) / 1000 /10

    def convert_to_json_format(self,raw_data,time_offset=0.0): 
        len2 = len(raw_data[1].split()[1:])
        list_data = []
        cur_list_data = []
        cunt = 0
        for cunt in range(0,len2):
            for rr in raw_data:
                if not rr.startswith("NOC"):
                    continue
                r = rr.split()[1:]
                if cunt == 0 :
                    timestamp = round((float(r[0]) - time_offset) * 1000, )
                    if timestamp < 0:
                        continue
                    cur_list_data.append(timestamp)
                else:
                    cur_list_data.append(round(self.toMB_s(r[cunt]), 2))
            list_data.append(cur_list_data)
            cur_list_data = []
        return list_data

    def create_json_file(self,json_data):
       
        ddrmc_name = []
        ddrmc_data = {}
        data = {
                "ddrmc_nsu_timestamp": json_data[0]
        }
        for k in range (len(self.base_addr)):
            ddrmc_name.append("ddrmc" + "@" + str(hex(self.base_addr[k])) + "_read")
            ddrmc_name.append("ddrmc" + "@" + str(hex(self.base_addr[k])) + "_write")

        for i in range (len(self.base_addr) * 2):# every addr map read and write
            ddrmc_data[ddrmc_name[i]] =  json_data[i+1]
        merge(data, ddrmc_data) 
        json_data = json.dumps(data, indent=4, separators=(",", ":"))
        with open('ddrmc_noc_nsu_data.json','w') as json_file:
            json_file.write(json_data)
        

if __name__ == '__main__':

    noc = NOC()
    noc.start(0.001)
    time.sleep(1)
    noc.stop()
    noc.transTimebase()
    raw_data = noc.getData()
    data_cur = noc.convert_to_json_format(raw_data)
    noc.create_json_file(data_cur)




#!/usr/bin/python3
# -*- coding:utf-8 -*-

# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ctypes import *
import logging

import tracer.tracerBase
import copy
import os

def record_data_factory(size=10):
    class APMRecord(Structure):
        _fields_ = [('time', c_double), ('data', c_longlong * size)]

        def output(self):
            data_cur = []
            list_data = []
            data_cur = [*self.data]
            list_data.append("NOC_NMU")
            list_data.append(self.time)
            for i in range(0, size):
                list_data.append(data_cur[i])
            return list_data
    data =  APMRecord()
    return data

class NMU:
    def __init__(self, options, path='/usr/lib/libnmuperf.so', enable=True):
        self.nmuLib = cdll.LoadLibrary(path)
        self.data = []
        self.base_addr = []
        self.location_name = []
        self.interval = 0
        self.npi_freq = 0
        self.act_period = 0
        self.record_data_len = 0
        self.started = False
        self.type = "noc_nmu"
        self.enabled = enable
        self.npi_freq = options.get("npi_freq",{})
        nodes = options.get("nodes", {})
        self.base_addr = self.get_noc_nmu_addr(nodes)
        self.npi_base_addr = self.get_noc_npi_addr(nodes)
        self.noc_nmu_addr = (c_int * len(self.base_addr))(*self.base_addr)
        self.nmuLib.create_noc_nmu_instance(self.type.encode(),int(self.npi_freq),self.noc_nmu_addr,self.npi_base_addr,int(len(self.base_addr)))

        

    def start(self, interval=1.0):
        if self.enabled == False:
            return
        self.interval = interval
        self.nmuLib.start.restype = c_int
        return self.nmuLib.start(c_double(interval))

    def pushData(self, data):
        if self.enabled == False:
            return
        for i in range(0, len(data.data)):
            data.data[i] = data.data[i]

        self.data.append(copy.deepcopy(data))


    def get_noc_npi_addr(self, noc_nodes):
        for k in range(len(noc_nodes)):
            if("npi" == noc_nodes[k].get("type", {})):
                return int(noc_nodes[k].get("phy_address", {}), 16)

    def get_noc_nmu_addr(self, noc_nodes):
        noc_nmu_addr = []
        for k in range(len(noc_nodes)):
            if("nmu" == noc_nodes[k].get("type", {})):
                addr = noc_nodes[k].get("phy_address", {})
                noc_nmu_addr.append(int(addr, 16))
                self.location_name.append(noc_nodes[k].get("location_name", {}))
        if len(noc_nmu_addr) > 8:
            logging.error("The maximum of sample channals is 8!")
            exit(1)
        return noc_nmu_addr

    def stop(self):
        if self.enabled == False:
            return
        self.nmuLib.stop()

        data = record_data_factory(len(self.base_addr) * 2)
        pd = pointer(data)

        while (self.nmuLib.pop_data(pd) == 0):
            self.data.append(copy.deepcopy(data))

        self.nmuLib.get_act_period.restype = c_double
        self.act_period = self.nmuLib.get_act_period()

        self.nmuLib.get_record_data_len.restype = c_int
        self.record_data_len = self.nmuLib.get_record_data_len()


    def transTimebase(self):
        for i in range(0, len(self.data)):
            for j in range(0, self.record_data_len):
                self.data[i].data[j] = int(
                    self.data[i].data[j] / self.act_period)


class nmuTracer(tracer.tracerBase.Tracer):
    def __init__(self):
        super().__init__('nmu', source=[],
                         compatible={'machine': ["aarch64"]})
        self.nmu_info = {}
        self.nmu = None
        self.nmu_enable = False
        self.nodes = []
        self.location_name = []

    def prepare(self, option: dict, debug: bool):
       nmuOption = option.get('tracer', {}).get('noc_perf', {})
       if nmuOption.get('noc_nmu_json_enable',{}) == "enable":
           self.interval = nmuOption.get("interval", 0)
           self.npi_freq = nmuOption.get("npi_freq",{})
           self.nodes = nmuOption.get("nodes", {})
           self.nmu = NMU(nmuOption)
           self.location_name = self.nmu.location_name
           self.nmu_info.update({"interval": self.interval})
           self.nmu_info.update({"npi_freq" : self.npi_freq})
           self.nmu_info.update({"nodes": self.nodes})
           self.nmu_info.update({"location_name": self.location_name})
           self.nmu_enable = True
       option['noc_nmu'] = self.nmu_enable
       return option

    def start(self):
        super().start()
        if self.nmu_enable:
            self.nmu.start(self.interval)

    def stop(self):
        super().stop()
        if self.nmu_enable:
            self.nmu.stop()

    def process(self, data, t_range=[]):
        if self.nmu_enable:
            self.nmu.transTimebase()
        pass

    def compatible(self, platform: {}):
        if super().compatible(platform) == False:
            return False
        if platform.get('model').startswith('xlnx,zocl-versal'):
            return True
        else:
            return False
    
    def getData(self):
        if self.nmu_enable:
            return self.nmu_info,[d.output() for d in self.nmu.data]
        else:
            print("noc_nmu feature disabled")

tracer.tracerBase.register(nmuTracer())



#!/usr/bin/env python

# Copyright 2022-2023 Advanced Micro Devices Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os.path
import xml.etree.ElementTree as ET

from python import Root

from functools import reduce


class mcInst:
    def __init__(self, inst):
        self.inst = inst

    def code_gen(self):
        inst = self.inst
        words = []
        reserved_id = 1
        reserved_name = "redvered"
        for w in inst.word_list:
            # uint32_t bank_addr : 14, bank_id : 6, dpby : 4, dpdon : 4, opcode : 4;
            f_code = []
            for f in w.field_list[::-1]:
                field_name = f.name
                if field_name == "reserved":
                    field_name = "%s_%d" % (reserved_name, reserved_id)
                    reserved_id += 1
                field_len = f.len
                f_code.append("{} : {}".format(field_name, field_len))

            f_code_str = reduce(lambda _x, _y: "{}, {}".format(_x, _y), f_code)
            words.append("uint32_t {};\n".format(f_code_str))
        code = "struct {} {{{}}};\n\n".format(
            inst.name, reduce(lambda _x, _y: _x + _y, words))

        return code


class InstTable:
    def __init__(self, insts):
        self.inst_table = insts

    def code_gen(self):
        code = "std::vector<class inst_desc> inst_table = {{{}}};\n"
        tmp = []
        for i in self.inst_table:
            ii = {"inst_name": i.name.upper(), "inst_opcode": i.opcode_str,
                  "inst_len": i.word_num}
            tmp.append(
                "create_inst_desc ({inst_name}, {inst_opcode}, {inst_len})".format(**ii))
        code = code.format(reduce(lambda _x, _y: "{},{}".format(_x, _y), tmp))

        return code


def gen_mc_header(root_list, dir="./"):

    for dpuInstance in root_list:
        h_f = "%s.h" % dpuInstance.version
        h = open(os.path.join(dir, h_f), "w+t")
        i_code = ""
        for i in dpuInstance.inst_list:
            i_code = mcInst(i).code_gen()
            h.write(i_code)

        t_code = InstTable(dpuInstance.inst_list).code_gen()
        h.write(t_code)


def main(wrk_dir="./"):
    xml_list = []
    for roots, dirs, files in os.walk("./xml"):
        for f in files:
            xml_list.append(os.path.join(roots, f))
    root_list = [Root.Root(ET.parse(x).getroot()) for x in xml_list]
    gen_mc_header(root_list)


if __name__ == "__main__":
    main()




# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from writer.parser.timelineUtil import *
from writer.parser.tracepointUtil import *
from writer.parser.statUtil import *
from writer.parser.ftraceUtil import *


def cppFuncPaired(start: vaiTimelineEvent, end: vaiTimelineEvent):
    if not (start.type == 'start' and end.type == 'done'):
        return None
    if end.ts < start.ts:
        return None
    if start.pid != end.pid:
        return None
    if start.code != end.code:
        return None

    pairedEvent = vaiTimelineEvent(
        start.startTime, start.pid, start.type, start.coreType, start.coreId)

    pairedEvent.type = 'period'
    pairedEvent.startTime = start.ts
    pairedEvent.endTime = end.ts
    pairedEvent.duration = end.ts - start.ts
    pairedEvent.code = start.code

    return pairedEvent


def convert_cppfunc(cppfunc_data):
    threads = {}
    pairedEvents = []
    for l in cppfunc_data:
        event = ftraceParse(l, {}).toTimelineEvent()
        pid = event.pid
        threads.setdefault(pid, []).append(event)

    stack = []
    for pid in threads.keys():
        for e in threads[pid]:
            stack.append(e)
            if len(stack) < 2:
                continue
            for s in stack[::-1]:
                r = cppFuncPaired(s, e)
                if r != None:
                    try:
                        stack.remove(e)
                        stack.remove(s)
                    except BaseException:
                        pass

                    pairedEvents.append(r)
                    break

    cpp_func_calls = statTable("C/C++ Function")
    for event in pairedEvents:
        if event.type != 'period':
            continue
        cpp_func_calls.add(event.code, event.duration)
    return cpp_func_calls.output()



# Copyright 2022-2023 Advanced Micro Devices Inc.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
import logging
import json
import uuid

from vaitraceOptions import merge

from writer.parser.vart_parser import *
from writer.parser.tracepointUtil import *
from writer.parser.statUtil import *
from writer.parser.pyfunc import *
from writer.parser.cppfunc import *
from writer.parser.cpu_task import *
from writer.parser.power import *

from writer.parser.sched_parser.sched_parser import sched_parser_main

vtf_events = []
dpu_ip_summary = {}


def convert_dpu(raw_data, dpu_ip_data, options):
    global dpu_ip_summary

    runmachine = options.get('control', {}).get('platform',{}).get('machine',{})
    dpu_parser = DPUEventParser()
    timelines = dpu_parser.parse(raw_data, dpu_ip_data, {})

    dpu_ip_summary = {"dpu": {"ip": dpu_parser.get_dpu_ip_summary()}}
    report = {
        "dpu": {
            "profiling": {},
            "scheduling": {}
        }
    }

    for timeline in timelines:
        if timeline.coreType == "DPU":
            cu_name = timeline.get_core_name()
            scheduling_rate = timeline.get_util()
            if cu_name == None or scheduling_rate == -1:
                continue

            # print(f"{dpu_name}: {scheduling_rate}")
            report["dpu"]["scheduling"].setdefault(cu_name, 0)
            report["dpu"]["scheduling"][cu_name] = scheduling_rate

    # "{subg_name},{runs},{cu_name},{batch},{min_t},{avg_t},{max_t},{workload},{effic},{read_wb_size},{read_fm_size},{write_fm_size},{mem_io_bw},{hw_rt},{depth},\n"
    dpu_summary = dpu_parser.get_dpu_profile_summary({"run_machine": runmachine},"json")

    for row in dpu_summary:
        row = row.strip().split(',')
        dpu_model_profiling_item = {
            "subg_name":        str(row[0]),
            "runs":             int(row[1]),
            "cu_name":          str(row[2]),
            "batch":            int(row[3]),
            "min_time":         float(row[4]),
            "avg_time":         float(row[5]),
            "max_time":         float(row[6]),
            "workload":         int(row[7]),
            "efficiency":       float(row[8]),
            "read_weight_bias_size":    int(row[9]),
            "read_feature_map_size":    int(row[10]),
            "write_feature_map_size":   int(row[11]),
            "mem_io_bw":        float(row[12]),
            "hardware_time":    float(row[13]),
            "depth":     int(row[14]),
            "input_tensor_shape": str(row[15].replace("_", ",")),
            "output_tensor_shape": str(row[16].replace("_", ",")),
            "run_performance" : float(row[17]),
            "inst_size": int(row[18])
        }

        # extended result
        dpu_model_profiling_item["feature_map_ave_rd_bandwidth"] = dpu_model_profiling_item["read_feature_map_size"] / \
                                (dpu_model_profiling_item["hardware_time"] / 1000)
        dpu_model_profiling_item["feature_map_ave_wr_bandwidth"] = dpu_model_profiling_item["write_feature_map_size"] / \
                                (dpu_model_profiling_item["hardware_time"] / 1000)

        # sort by cu
        cu_name = dpu_model_profiling_item["cu_name"]
        report["dpu"]["profiling"].setdefault(cu_name, [])
        report["dpu"]["profiling"][cu_name].append(dpu_model_profiling_item)

    return report


def convert_power_info(raw_data):
    return analyse_power(raw_data)


def get_power_info(power_info_data):
    peak = power_info_data.get("peak_power", 0)
    idle = power_info_data.get("idle_power", 0)
    average = power_info_data.get("ave_power", 0)
    limit = power_info_data.get("limit_power", -1)

    return {"board": {
        "power.peak": peak,
        "power.idle": idle,
        "power.average": average,
        "power.limit": limit
    }
    }

def toMB_s(r):

    return float(r) / 1000 / 1000

def convert_xapm(apm_data,option={}):
    apm_events = []

    for rr in apm_data:
        if not rr.startswith("APM "):
            continue
        r = rr.split()[1:]

        s_1_r = round(toMB_s(r[1]), 2)
        s_1_w = round(toMB_s(r[2]), 2)
        s_2_r = round(toMB_s(r[3]), 2)
        s_2_w = round(toMB_s(r[4]), 2)
        s_3_r = round(toMB_s(r[5]), 2)
        s_3_w = round(toMB_s(r[6]), 2)
        s_4_r = round(toMB_s(r[7]), 2)
        s_4_w = round(toMB_s(r[8]), 2)
        s_5_r = round(toMB_s(r[9]), 2)
        s_5_w = round(toMB_s(r[10]), 2)

        total_value = (s_1_r + s_1_w + s_2_r + s_2_w + s_3_r + s_3_w +
                s_4_r + s_4_w + s_5_r +s_5_w)
        apm_events.append(total_value)

    return apm_events

def convert_ddrmc_nsu(raw_data,option={},time_offset=0.0):
    ddrmc_type = option.get('ddrmc_json_enable', {})
    len2 = len(raw_data[1].split()[1:])
    list_data = []
    cur_list_data = []
    ddrmc_base_addr = []
    cunt = 0
    for cunt in range(0,len2):
        for rr in raw_data:
            if not rr.startswith("APM"):
                continue
            r = rr.split()[1:]
            if cunt == 0 :
                timestamp = round((float(r[0]) - time_offset) * 1000, )
                if timestamp < 0:
                    continue
                cur_list_data.append(timestamp)
            else:
                cur_list_data.append(round(toMB_s(r[cunt]), 2))
        list_data.append(cur_list_data)
        cur_list_data = []
    ddrmc_base_addr = option.get("ddrmc_base_addr", {})
    if ddrmc_type == "enable": 
        ddrmc_nsu_json_file(list_data, ddrmc_base_addr)

def ddrmc_nsu_json_file(json_data, base_addr):
    ddrmc_name = []
    ddrmc_data = {}
    data = {
            "ddrmc_nsu_timestamp": json_data[0]
    }
    for k in range (len(base_addr)):
        ddrmc_name.append("ddrmc" + "@" + str(hex(base_addr[k])) + "_read")
        ddrmc_name.append("ddrmc" + "@" + str(hex(base_addr[k])) + "_write")

    for i in range (0,len(base_addr) * 2):# every addr map read and write
        ddrmc_data[ddrmc_name[i]] =  json_data[i+1]
    merge(data, ddrmc_data)
    json_data = json.dumps(data, indent=4, separators=(",", ":"))
    with open('ddrmc_noc_nsu_data.json','w') as json_file:
        json_file.write(json_data)

def convert_noc_nmu(nmu_data,options,time_offset=0):

    print(nmu_data)
    if options.get('noc_nmu', '') != True:
        return
    sample_data = []
    nmu_info = nmu_data[0]
    sample_data = nmu_data[1]
    json_data = nmu_convert_data_to_json(sample_data)
    location_name = nmu_info.get("location_name", {})
    create_noc_nmu_json_file(json_data, location_name)

def nmu_convert_data_to_json(raw_data, time_offset=0.0):
    len2 = len(raw_data[1]) - 1
    list_data = []
    cur_list_data = []
    cunt = 0
    for cunt in range(0,len2):
        for rr in raw_data:
            if rr[0] != "NOC_NMU":
                continue
            r = rr[1:]
            if cunt == 0 :
                timestamp = round((float(r[0]) - time_offset) * 1000, )
                if timestamp < 0:
                    continue
                cur_list_data.append(timestamp)
            else:
                cur_list_data.append(round(toMB_s(r[cunt]), 2))
        list_data.append(cur_list_data)
        cur_list_data = []
    return list_data

def create_noc_nmu_json_file(json_data, location_name):
    json_name = []
    D = {}
    data = {
            "noc_nmu_timestamp": json_data[0]
    }
    for k in range(0, len(location_name)):
        json_name.append(location_name[k] + "_READ")
        json_name.append(location_name[k] + "_WRITE")
    for i in range(0,len(json_name)):
        D[json_name[i]] = json_data[i+1]
    merge(data, D)
    json_data = json.dumps(data, indent=4, separators=(",", ":"))
    with open('noc_nmu_data.json','w') as json_file:
        json_file.write(json_data)


def get_peak_mem_io(mem_io_summary):

    peak_mem = round(max(mem_io_summary), 2)
    ave_mem = round((sum(mem_io_summary) / len(mem_io_summary)), 2)
    return {"mem_io": {
        "peak_mem_io" : str(peak_mem) + "MB/s",
        "average_mem_io" : str(ave_mem) + "MB/s"
    }
    }

def get_cpu_task_info(cpu_task_summary):
    report = {
        "cpu": {
            "graph_runner.cpu_tasks": [],
            "onnx_runner.cpu_tasks": []
        }
    }

    for row in cpu_task_summary:
        items = row.strip().split(',')

        subgraph_name = items[0].replace("subgraph_", "", 1)

        runs = int(items[1])
        min_t = float(items[3])
        ave_t = float(items[4])
        max_t = float(items[5])
        ops = items[6]

        report_item = {
            "subgraph": subgraph_name,
            "ops": ops,
            "average_time": ave_t,
            "min_time": min_t,
            "max_time": max_t,
            "runs": runs
        }

        report["cpu"]["graph_runner.cpu_tasks"].append(report_item)

    return report


def get_cpu_func_info(cpp_summary, py_summary):
    report = {
        "cpu": {
            "vitis_ai_libaray.cpu_functions": []
        }
    }

    for row in cpp_summary + py_summary:
        items = row.strip().split(',')

        if row in py_summary:
            function_name = "%s@py" % items[0]
        else:
            function_name = items[0]
        runs = int(items[1])

        min_t = float(items[3])
        ave_t = float(items[4])
        max_t = float(items[5])

        report_item = {
            "func": function_name,
            "average_time": ave_t,
            "min_time": min_t,
            "max_time": max_t,
            "runs": runs
        }

        report["cpu"]["vitis_ai_libaray.cpu_functions"].append(report_item)

    return report


def get_cmd_info(cmd_data):
    report = {
        "vaitrace_environment": {}
    }

    report["vaitrace_environment"]["cmd"] = cmd_data.get("cmd", [])
    report["vaitrace_environment"]["platform"] = cmd_data.get("platform", [])
    report["vaitrace_environment"]["time"] = cmd_data.get("time", [])

    return report


def xat_to_json(xat, options):
    global dpu_ip_summary

    saveTo = options.get('cmdline_args', {}).get('output', None)

    convert_ddrmc_nsu(xat.get('xapm',{}),options)
    convert_noc_nmu(xat.get('nmu',{}), options)
    dpuProfilingSummary = convert_dpu(xat.get('vart'), xat.get('hwInfo'), options)

    powerInfoSummary = convert_power_info(xat.get('power', {}))
    cpuTaskSummary = convert_cpu_task(xat.get('vart', {}))
    pyFuncSummary = convert_pyfunc(xat.get('pyfunc', {}))
    cppFuncSummary = convert_cppfunc(xat.get('function', {}))
    memIoSummary = convert_xapm(xat.get('xapm', {}))

    cpuUtilSummary = sched_parser_main(xat.get('sched', {}))

    output_f = sys.stdout
    report = {"uuid": uuid.uuid1().hex}

    if saveTo != None:
        logging.info("Saving report to: %s" % saveTo)
        try:
            output_f = open(saveTo, "w+t")
        except:
            output_f = sys.stdout
            logging.error("Opening file failed: %s" % saveTo)

    merge(report, dpuProfilingSummary)
    merge(report, get_power_info(powerInfoSummary))
    merge(report, get_cpu_task_info(cpuTaskSummary))
    merge(report, get_cpu_func_info(cppFuncSummary, pyFuncSummary))
    merge(report, get_peak_mem_io(memIoSummary))
    merge(report, get_cmd_info(xat.get('cmd', {})))
    merge(report, dpu_ip_summary)
    merge(report, cpuUtilSummary)

    print(json.dumps(report, indent=4, separators=(",", ":")), file=output_f)
